use std::ops::Mul;

use enum_iterator::Sequence;
use palette::Oklch;
use serde::{Deserialize, Serialize};
use strum::Display;
use wasm_bindgen::prelude::wasm_bindgen;

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
#[wasm_bindgen]
pub struct ColorShifting {
    pub chroma: f32,
    pub lightness: f32,
}

#[wasm_bindgen]
impl ColorShifting {
    #[wasm_bindgen(constructor)]
    pub fn new(chroma: f32, lightness: f32) -> Self {
        ColorShifting { chroma, lightness }
    }
}

impl Mul<ColorShifting> for Oklch<f32> {
    type Output = Oklch<f32>;

    fn mul(self, rhs: ColorShifting) -> Self::Output {
        Oklch::new(
            self.l
                + if rhs.lightness > 0.0 {
                    (1.0 - self.l) * rhs.lightness
                } else {
                    self.l * rhs.lightness * -1.0
                },
            self.chroma
                + if rhs.chroma > 0.0 {
                    (100.0 - self.chroma) * rhs.chroma
                } else {
                    -(self.chroma * rhs.chroma)
                },
            self.hue,
        )
    }
}

impl Mul<ColorShifting> for &Oklch<f32> {
    type Output = Oklch<f32>;

    fn mul(self, rhs: ColorShifting) -> Self::Output {
        Oklch::new(
            self.l
                + if rhs.lightness > 0.0 {
                    (1.0 - self.l) * rhs.lightness
                } else {
                    self.l * rhs.lightness * -1.0
                },
            self.chroma
                + if rhs.chroma > 0.0 {
                    (100.0 - self.chroma) * rhs.chroma
                } else {
                    -(self.chroma * rhs.chroma)
                },
            self.hue,
        )
    }
}

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
#[wasm_bindgen]
pub struct SchemeSetting {
    pub hover: ColorShifting,
    pub active: ColorShifting,
    pub focus: ColorShifting,
    pub disabled: ColorShifting,
    pub dark_convert: ColorShifting,
    pub expand_method: ColorExpand,
    pub wacg_follows: WACGSetting,
}

#[derive(Debug, Clone, Copy, Display, Sequence, Serialize, Deserialize)]
#[wasm_bindgen]
#[repr(u8)]
pub enum ColorExpand {
    Complementary,
    Analogous,
    AnalogousAndComplementary,
    Triadic,
    SplitComplementary,
    Tetradic,
    Square,
}

impl ColorExpand {
    pub fn label(&self) -> String {
        match self {
            ColorExpand::Complementary => "Complementary".to_string(),
            ColorExpand::Analogous => "Analogous".to_string(),
            ColorExpand::AnalogousAndComplementary => "Analogous and Complementary".to_string(),
            ColorExpand::Triadic => "Triadic".to_string(),
            ColorExpand::SplitComplementary => "Split Complementary".to_string(),
            ColorExpand::Tetradic => "Tetradic".to_string(),
            ColorExpand::Square => "Square".to_string(),
        }
    }
}

#[derive(Debug, Clone, Copy, Display, Sequence, Serialize, Deserialize)]
#[wasm_bindgen]
#[repr(u8)]
pub enum WACGSetting {
    Fixed,
    AutomaticAA,
    AutomaticAAA,
}

impl WACGSetting {
    pub fn label(&self) -> String {
        match self {
            WACGSetting::Fixed => "Fixed".to_string(),
            WACGSetting::AutomaticAA => "Automatic AA".to_string(),
            WACGSetting::AutomaticAAA => "Automatic AAA".to_string(),
        }
    }
}

impl Default for SchemeSetting {
    fn default() -> Self {
        SchemeSetting {
            hover: ColorShifting {
                chroma: 0.0,
                lightness: 0.3,
            },
            active: ColorShifting {
                chroma: 0.0,
                lightness: -0.2,
            },
            focus: ColorShifting {
                chroma: 0.0,
                lightness: 0.5,
            },
            disabled: ColorShifting {
                chroma: -0.9,
                lightness: -0.2,
            },
            dark_convert: ColorShifting {
                chroma: -0.3,
                lightness: -0.3,
            },
            expand_method: ColorExpand::AnalogousAndComplementary,
            wacg_follows: WACGSetting::AutomaticAAA,
        }
    }
}

#[wasm_bindgen]
impl SchemeSetting {
    #[wasm_bindgen(constructor)]
    pub fn new(
        hover: ColorShifting,
        active: ColorShifting,
        focus: ColorShifting,
        disabled: ColorShifting,
        dark_convert: ColorShifting,
        expand_method: ColorExpand,
        wacg_follows: WACGSetting,
    ) -> Self {
        SchemeSetting {
            hover,
            active,
            focus,
            disabled,
            dark_convert,
            expand_method,
            wacg_follows,
        }
    }
}
